import os
import unittest

from tinydb import Query

from tosker.graph.nodes import Container
from tosker.graph.template import Template
from tosker.storage import Memory, Storage


class Test_Storage(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        pass

    def setUp(self):
        Storage.set_db('/tmp')

    def tearDown(self):
        os.remove('/tmp/db.json')

    def test_storage_insert(self):
        res = Storage.insert({'name': 'hello'})
        res = Storage.search(Query().name == 'hello')
        self.assertEqual(len(res), 1)

    def test_storage_update(self):
        self.test_storage_insert()

        res = Storage.update({'name': 'hello2'}, Query().name == 'hello')
        self.assertEqual(len(res), 1)

        res = Storage.search(Query().name == 'hello2')
        self.assertEqual(len(res), 1)

    def test_storage_all(self):
        self.test_storage_insert()

        res = Storage.all()
        self.assertEqual(len(res), 1)

    def test_memory_update_state(self):
        cont = Container('container_test')
        cont.tpl = Template('template_test')

        Memory.update_state(cont, Memory.STATE.CREATED)
        state = Memory.get_comp_state(cont)
        self.assertEqual(state, Memory.STATE.CREATED)

        Memory.update_state(cont, Memory.STATE.STARTED)
        state = Memory.get_comp_state(cont)
        self.assertEqual(state, Memory.STATE.STARTED)

        Memory.update_state(cont, Memory.STATE.CREATED)
        state = Memory.get_comp_state(cont)
        self.assertEqual(state, Memory.STATE.CREATED)

        Memory.update_state(cont, Memory.STATE.DELETED)
        state = Memory.get_comp_state(cont)
        self.assertEqual(state, Memory.STATE.DELETED)

    def test_memory_get_comps(self):
        tpl = Template('template_test')
        cont1 = Container('container_test1')
        cont1.tpl = tpl
        cont2 = Container('container_test2')
        cont2.tpl = tpl
        cont3 = Container('container_test3')
        cont3.tpl = tpl

        Memory.update_state(cont1, Memory.STATE.CREATED)
        Memory.update_state(cont2, Memory.STATE.STARTED)
        Memory.update_state(cont3, Memory.STATE.DELETED)

        comps = Memory.get_comps()
        self.assertEqual(len(comps), 3)

        comps = Memory.get_comps(filters={'state': Memory.STATE.CREATED})
        self.assertEqual(len(comps), 1)
        self.assertEqual(comps[0]['full_name'],
                         'template_test.container_test1')

        comps = Memory.get_comps(filters={'state': Memory.STATE.STARTED})
        self.assertEqual(len(comps), 1)
        self.assertEqual(comps[0]['full_name'],
                         'template_test.container_test2')



import unittest

from .test_tosca_base import Test_Orchestrator


class Test_Thoughts(Test_Orchestrator):

    def setUp(self):
        super(self.__class__, self).setUp()
        self.file = 'data/examples/thoughts-app/thoughts.csar'

    def test(self):
        self.create()
        self.start()
        self.stop()
        self.start()
        self.stop()
        self.delete()


if __name__ == '__main__':
    unittest.main()



import unittest

from tosker import docker_interface as docker
from tosker.graph.artifacts import DockerImageExecutable
from tosker.graph.nodes import Container
from tosker.graph.template import Template

# from tosker.helper import Logger


class Test_DockerInterface(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # Logger.set(helper.get_consol_handler(), False)
        docker.remove_all_containers()

        self._container = Container('test_container')
        self._container.image = DockerImageExecutable('alpine')
        self._container.cmd = 'ping 127.0.0.1'
        tpl = Template('test_app')
        tpl.tmp_dir = '/tmp/tosker_tests'
        self._container.tpl = tpl

        docker.create_network(self._container)

    @classmethod
    def tearDownClass(self):
        docker.delete_network(self._container)

    def setUp(self):
        pass

    def tearDown(self):
        docker.remove_all_containers()

    def test_create_container(self):
        docker.create_container(self._container, force=False)
        self.assertIsNotNone(
            docker.inspect_container(self._container)
        )
        self.assertRaises(
            Exception,
            lambda: docker.create_container(self._container, force=False)
        )
        self.assertIsNotNone(
            docker.inspect_container(self._container)
        )

    def test_start_container(self):
        docker.create_container(self._container)
        self.assertIsNotNone(
            docker.inspect_container(self._container)
        )
        docker.start_container(self._container)
        stat = docker.inspect_container(self._container)
        self.assertIsNotNone(stat)
        self.assertTrue(stat['State']['Running'])

    def test_start_container_error(self):
        self.assertRaises(
            Exception,
            lambda: docker.start_container(self._container)
        )
        self.assertIsNone(
            docker.inspect_container(self._container)
        )

    def test_stop_container(self):
        docker.create_container(self._container)
        self.assertIsNotNone(
            docker.inspect_container(self._container)
        )
        docker.start_container(self._container)

        stat = docker.inspect_container(self._container)
        self.assertIsNotNone(stat)
        self.assertTrue(stat['State']['Running'])

        docker.stop_container(self._container)
        stat = docker.inspect_container(self._container)
        self.assertIsNotNone(stat)
        self.assertFalse(stat['State']['Running'])

    def test_stop_container_error(self):
        self.assertRaises(
            Exception,
            lambda: docker.stop_container(self._container)
        )
        self.assertIsNone(
            docker.inspect_container(self._container)
        )

    def test_container_delete(self):
        docker.create_container(self._container)
        self.assertIsNotNone(
            docker.inspect_container(self._container)
        )

        docker.delete_container(self._container)
        self.assertIsNone(
            docker.inspect_container(self._container)
        )

    def test_container_delete_error(self):
        self.assertRaises(
            Exception,
            lambda: docker.delete_container(self._container)
        )
        self.assertIsNone(
            docker.inspect_container(self._container)
        )

    def test_container_exec(self):
        docker.create_container(self._container)
        self.assertIsNotNone(
            docker.inspect_container(self._container)
        )
        docker.start_container(self._container)
        docker.exec_cmd(self._container, 'echo hello!')

    def test_container_exec_error(self):
        docker.create_container(self._container)
        self.assertIsNotNone(
            docker.inspect_container(self._container)
        )
        self.assertRaises(
            Exception,
            lambda: docker.exec_cmd(self._container, 'echo hello!')
        )
        docker.delete_container(self._container)
        self.assertIsNone(
            docker.inspect_container(self._container)
        )
        self.assertRaises(
            Exception,
            lambda: docker.exec_cmd(self._container, 'echo hello!')
        )

    # def test_network(self):
    #     docker.delete_network()
    #     docker.delete_network()
    #     # TODO: assert that the network is correctlly created
    #     docker.create_network()
    #     docker.create_network()



import unittest

# from tosker.graph.template import Template
# from tosker.graph.nodes import Container, Software, Volume
# from tosker.tosca_parser import get_tosca_template


class Test_Graph(unittest.TestCase):
    pass
    # TODO: the test must target the orchestrator

    # def _assert_sorting(self, tpl):
    #     running = set()
    #     for c in tpl.nodes:
    #         for r in c.relationships:
    #             self.assertIn(r.to.name, running)
    #         running.add(c.name)

    # def test_deployment_order(self):
    #     tpl = get_tosca_template(
    #         'tosker/tests/TOSCA/software-link/software.yaml')
    #     self._assert_sorting(tpl)
    #
    # def test_components_1(self):
    #     tpl = get_tosca_template(
    #         'tosker/tests/TOSCA/software-link/software.yaml',
    #         components=['server1'])
    #
    #     self._assert_sorting(tpl)
    #     self.assertEqual(len(tpl.nodes), 2)
    #     for c in tpl.nodes:
    #         self.assertIn(c.name, ('server1', 'nodejs1'))
    #
    # def test_components_2(self):
    #     tpl = get_tosca_template(
    #         'tosker/tests/TOSCA/software-link/software.yaml',
    #         components=['server2'])
    #
    #     self._assert_sorting(tpl)
    #     self.assertEqual(len(tpl.nodes), 2)
    #     for c in tpl.nodes:
    #         self.assertIn(c.name, ('server2', 'nodejs2'))
    #
    # def test_components_3(self):
    #     tpl = get_tosca_template(
    #         'tosker/tests/TOSCA/software-link/software.yaml',
    #         components=['server3'])
    #
    #     self._assert_sorting(tpl)
    #     self.assertEqual(len(tpl.nodes), 6)
    #     for c in tpl.nodes:
    #         self.assertIn(c.name, ('server1', 'nodejs1',
    #                                'server2', 'nodejs2',
    #                                'server3', 'nodejs3'))
    #
    # def test_components_1_2(self):
    #     tpl = get_tosca_template(
    #         'tosker/tests/TOSCA/software-link/software.yaml',
    #         components=['server1', 'server2'])
    #
    #     self._assert_sorting(tpl)
    #     self.assertEqual(len(tpl.nodes), 4)
    #     for c in tpl.nodes:
    #         self.assertIn(c.name, ('server1', 'nodejs1',
    #                                'server2', 'nodejs2'))
    #
    # def test_components_1_3(self):
    #     tpl = get_tosca_template(
    #         'tosker/tests/TOSCA/software-link/software.yaml',
    #         components=['server1', 'server3'])
    #     self._assert_sorting(tpl)
    #     self.assertEqual(len(tpl.nodes), 6)
    #     for c in tpl.nodes:
    #         self.assertIn(c.name, ('server1', 'nodejs1',
    #                                'server2', 'nodejs2',
    #                                'server3', 'nodejs3'))
    #
    # def test_components_2_3(self):
    #     tpl = get_tosca_template(
    #         'tosker/tests/TOSCA/software-link/software.yaml',
    #         components=['server2', 'server3'])
    #
    #     self._assert_sorting(tpl)
    #     self.assertEqual(len(tpl.nodes), 6)
    #     for c in tpl.nodes:
    #         self.assertIn(c.name, ('server1', 'nodejs1',
    #                                'server2', 'nodejs2',
    #                                'server3', 'nodejs3'))
    #
    # def test_components_n1(self):
    #     tpl = get_tosca_template(
    #         'tosker/tests/TOSCA/software-link/software.yaml',
    #         components=['nodejs1'])
    #
    #     self._assert_sorting(tpl)
    #     self.assertEqual(len(tpl.nodes), 1)
    #     for c in tpl.nodes:
    #         self.assertIn(c.name, ('nodejs1'))
    #
    # def test_components_n2(self):
    #     tpl = get_tosca_template(
    #         'tosker/tests/TOSCA/software-link/software.yaml',
    #         components=['nodejs2'])
    #
    #     self._assert_sorting(tpl)
    #     self.assertEqual(len(tpl.nodes), 1)
    #     for c in tpl.nodes:
    #         self.assertIn(c.name, ('nodejs2'))
    #
    # def test_components_n3(self):
    #     tpl = get_tosca_template(
    #         'tosker/tests/TOSCA/software-link/software.yaml',
    #         components=['nodejs3'])
    #
    #     self._assert_sorting(tpl)
    #     self.assertEqual(len(tpl.nodes), 1)
    #     for c in tpl.nodes:
    #         self.assertIn(c.name, ('nodejs3'))



'''
Volume manager module
'''
# from ..helper import Logger
from .. import docker_interface
from ..graph.nodes import Volume


class Volume_manager:

    @staticmethod
    def create(node):
        assert isinstance(node, Volume)
        docker_interface.create_volume(node)

    # @staticmethod
    # def delete(self, node):
    #     assert isinstance(node, Volume)
    #     docker_interface.delete_volume(node)



import sys
import unittest
from contextlib import contextmanager

from six import StringIO

from tosker import docker_interface as docker
from tosker.orchestrator import Orchestrator

from .test_tosca_base import Test_Orchestrator


@contextmanager
def redirect_stdout(new_target):
    old_target, sys.stdout = sys.stdout, new_target  # replace sys.stdout
    try:
        yield new_target  # run some code with the replaced stdout
    finally:
        sys.stdout = old_target  # restore to the previous value


class Test_Hello(Test_Orchestrator):

    def setUp(self):
        super(self.__class__, self).setUp()
        with redirect_stdout(StringIO()):
            self.orchestrator = Orchestrator(quiet=False)
        self.file = 'data/examples/hello.yaml'

    def test(self):
        with redirect_stdout(StringIO()):
            self.create()
            self.start_check_exit()
            self.stop()
            self.start_check_exit()

        # verify output
        temp_stdout = StringIO()
        with redirect_stdout(temp_stdout):
            self.stop()

        con_id = docker.inspect_container('hello.hello_container')['Id']

        output = temp_stdout.getvalue().strip()
        self.assertIn('- container_id: {}'.format(con_id), output)
        self.assertIn('- env_variable: Luca', output)

        with redirect_stdout(StringIO()):
            self.delete()


if __name__ == '__main__':
    unittest.main()



'''
Storage module
'''
from enum import Enum
from functools import reduce, wraps

import six
from tinydb import Query, TinyDB

from .graph.nodes import Root
from .helper import Logger


class Storage:
    def _singleton(func):
        @wraps(func)
        def func_wrapper(*args, **kwds):
            if Storage.db is not None:
                return func(*args, **kwds)
            else:
                raise Exception('First call "set_db" method')
        return func_wrapper

    @staticmethod
    def set_db(path):
        Storage.db = TinyDB('{}/db.json'.format(path))

    @staticmethod
    @_singleton
    def purge():
        return Storage.db.purge_tables()

    @staticmethod
    @_singleton
    def insert(obj):
        return Storage.db.insert(obj)

    @staticmethod
    @_singleton
    def search(condition):
        return Storage.db.search(condition)

    @staticmethod
    @_singleton
    def update(fields, condition):
        return Storage.db.update(fields, condition)

    @staticmethod
    @_singleton
    def remove(condition):
        return Storage.db.remove(condition)

    @staticmethod
    @_singleton
    def all():
        return Storage.db.all()


class Memory(Storage):
    class STATE(Enum):
        DELETED = 'deleted'
        CREATED = 'created'
        STARTED = 'started'

    @staticmethod
    def _comp_to_dict(comp):
        return {
            'name': comp.name,
            'full_name': comp.full_name,
            'type': type(comp).__name__,
            'app_name': comp.tpl.name,
            'state': 'deleted'
        }

    @staticmethod
    def update_state(comp, state):
        assert state in Memory.STATE

        def _update_state(full_name, state):
            log = Logger.get(__name__)
            log.debug('update {} to {}'.format(full_name, state))
            if (state == Memory.STATE.DELETED):
                return Memory.remove(Query().full_name == full_name)
            else:
                return Memory.update({'state': state.value},
                                     Query().full_name == full_name)

        if isinstance(comp, six.string_types):
            _update_state(comp, state)
        elif isinstance(comp, dict):
            res = _update_state(comp['full_name'], state)
            if len(res) < 1:
                # comp not found
                comp['state'] = state.value
                Memory.insert(comp)
        elif isinstance(comp, Root):
            res = _update_state(comp.full_name, state)
            if len(res) < 1:
                # comp not found
                comp_dict = Memory._comp_to_dict(comp)
                comp_dict['state'] = state.value
                Memory.insert(comp_dict)
        else:
            raise AssertionError()

    @staticmethod
    def insert(comp):
        assert isinstance(comp, (dict, Root))
        if isinstance(comp, Root):
            comp = Memory._comp_to_dict(comp)

        return Storage.insert(comp)

    @staticmethod
    def get_comp_state(comp):
        assert isinstance(comp, Root)
        res = Memory.search(Query().full_name == comp.full_name)
        return Memory.STATE(res[0]['state']) \
            if len(res) == 1 else Memory.STATE.DELETED

    @staticmethod
    def get_comps(app_name=None, filters={}):
        queries = []
        for k, v in filters.items():
            if isinstance(v, Memory.STATE):
                v = v.value
            queries.append(Query()[k] == v)

        if app_name is not None:
            queries.append(Query()['app_name'] == app_name)

        if len(queries) > 0:
            cond = reduce(lambda a, b: a & b, queries)
            return Memory.search(cond)
        else:
            return Memory.all()



'''
Tosca parser module
'''
import re
from os import path

import toscaparser
from toscaparser.prereq.csar import CSAR
from toscaparser.tosca_template import ToscaTemplate

from . import helper
from .graph.artifacts import (Dockerfile, DockerfileExecutable, DockerImage,
                              DockerImageExecutable, File)
from .graph.nodes import Container, Software, Volume
from .graph.template import Template
from .helper import Logger

_log = None

# CUSTOM TYPE
# PERSISTENT_CONTAINER = 'tosker.nodes.Container.Executable'
CONTAINER = 'tosker.nodes.Container'
VOLUME = 'tosker.nodes.Volume'
SOFTWARE = 'tosker.nodes.Software'
IMAGE = 'tosker.artifacts.Image'
IMAGE_EXE = 'tosker.artifacts.Image.Service'
DOCKERFILE = 'tosker.artifacts.Dockerfile'
DOCKERFILE_EXE = 'tosker.artifacts.Dockerfile.Service'

# EXECUTABLE_IMAGE = 'tosker.artifacts.ExecutableImage'
# IMAGE2 = 'tosca.artifacts.Deployment.Image.Container.Docker'
# DOCKERFILE2 = 'tosca.artifacts.File'

# REQUIREMENTS
CONNECT = 'connection'
DEPEND = 'dependency'
ATTACH = 'storage'
HOST = 'host'


# def _check_requirements(node, running):
#     for req in node.requirements:
#         for key, value in req.items():
#             value = value['node'] if type(value) is dict else value
#             if value not in running:
#                 return False
#     return True

# def _parse_path(base_path, value):
#     abs_path = path.abspath(
#         path.join(base_path, value)
#     )
#     split_path = abs_path.split('/')
#     return {'path': '/'.join(split_path[:-1]),
#             'file': split_path[-1],
#             'file_path': abs_path}


def _get_file(base_path, name, file):
    abs_path = path.abspath(
        path.join(base_path, file)
    )
    return File(name, abs_path)


def _parse_conf(tpl, node, repos, base_path):
    def _parse_map(m):
        res = {}
        for key, value in m.items():
            res[key] = value
        return res

    conf = None
    if node.type == CONTAINER:
        conf = Container(node.name)

        def parse_dockerfile(name, file, executable=False):
            # dockerfile = path.abspath(path.join(base_path, art['file']))
            conf.image = DockerfileExecutable(name, file) if executable else\
                         Dockerfile(name, file)

        def parse_image(name, repo, executable=False):
            conf.image = DockerImageExecutable(name) if executable else\
                         DockerImage(name)
            if repo:
                p = re.compile('(https://|http://)')
                repo = p.sub('', repos[repo]).strip('/')
                if repo != 'registry.hub.docker.com':
                    # TODO: test private repository
                    conf.image = '/'.join([repo.strip('/'),
                                           conf.image.format.strip('/')])

        # get artifacts
        artifacts = node.entity_tpl['artifacts']
        for key, value in artifacts.items():
            if isinstance(value, dict):
                name = value['file']
                art_type = value['type']
                repo = value.get('repository', None)

                if art_type == DOCKERFILE or art_type == DOCKERFILE_EXE:
                    dockerfile = path.abspath(
                                    path.join(base_path, name)
                                 )
                    _log.debug('dockerfile: {}'.format(dockerfile))
                    # if path.isfile(dockerfile) and repo is None:
                    _log.debug('Find a Dockerfile')
                    parse_dockerfile(key, dockerfile,  # .strip('/Dockerfile'),
                                     art_type == DOCKERFILE_EXE)
                elif art_type == IMAGE or art_type == IMAGE_EXE:
                    _log.debug('Find an Immage')
                    parse_image(name, repo, art_type == IMAGE_EXE)

        # get properties
        if 'properties' in node.entity_tpl:
            if 'env_variable' in node.entity_tpl['properties']:
                values = node.entity_tpl['properties']['env_variable']
                conf.env = _parse_map(values)

            if 'command' in node.entity_tpl['properties']:
                conf.cmd = node.entity_tpl['properties']['command']

            if 'ports' in node.entity_tpl['properties']:
                values = node.entity_tpl['properties']['ports']
                conf.ports = _parse_map(values)

            if 'share_data' in node.entity_tpl['properties']:
                values = node.entity_tpl['properties']['share_data']
                conf.share_data = _parse_map(values)

    elif node.type == VOLUME:
        conf = Volume(node.name)

    elif node.type == SOFTWARE:
        conf = Software(node.name)
        if 'artifacts' in node.entity_tpl:
            artifacts = node.entity_tpl['artifacts']
            for key, value in artifacts.items():
                _log.debug('artifacts: {}'.format(value))
                conf.add_artifact(_get_file(base_path, key, value))
                # TODO: parse also dictionary artifacts
                _log.debug('artifacts: {}'.format(conf.artifacts))

        # get interfaces
        if 'interfaces' in node.entity_tpl and \
                'Standard' in node.entity_tpl['interfaces']:
            interfaces = node.entity_tpl['interfaces']['Standard']
            intf = {}
            for key, value in interfaces.items():
                intf[key] = {}
                if 'implementation' in value:
                    abs_path = path.abspath(
                        path.join(base_path, value['implementation'])
                    )
                    # path_split = abs_path.split('/')

                    intf[key]['cmd'] = File(None, abs_path)
                    # {
                    #     'file': path_split[-1],
                    #     'path': '/'.join(path_split[:-1]),
                    #     'file_path': abs_path
                    # }
                    _log.debug('path: {} file: {}'
                               .format(intf[key]['cmd'].path,
                                       intf[key]['cmd'].file))
                if 'inputs' in value:
                    intf[key]['inputs'] = value['inputs']
                    # intf[key]['inputs'] = _parse_map(value['inputs'])

            conf.interfaces = intf

    else:
        raise ValueError(
            'node type "{}" not supported!'.format(node.type))

    # get requirements
    if 'requirements' in node.entity_tpl:
        requirements = node.entity_tpl['requirements']
        if requirements is not None:
            for value in requirements:
                if CONNECT in value:
                    conf.add_connection(value[CONNECT])
                if DEPEND in value:
                    conf.add_depend(value[DEPEND])
                if HOST in value:
                    if isinstance(value[HOST], dict):
                        conf.host = value[HOST]['node']
                    else:
                        conf.host = value[HOST]
                if ATTACH in value:
                    volume = value[ATTACH]
                    if isinstance(volume, dict):
                        conf.add_volume(volume['node'], volume['relationship']
                                                              ['properties']
                                                              ['location'])
    conf.tpl = tpl
    return conf


def get_tosca_template(file_path, inputs={}):
    global _log
    _log = Logger.get(__name__)

    # Work around bug validation csar of toscaparser
    if file_path.endswith(('.zip', '.csar')):
        csar = CSAR(file_path)
        try:
            csar.validate()
        except ValueError as e:
            _log.debug(e)
            if not str(e).startswith("The resource") or \
               not str(e).endswith("does not exist."):
                raise e

        csar.decompress()
        file_path = path.join(csar.temp_dir, csar.get_main_template())

    tosca = ToscaTemplate(file_path, inputs)

    base_path = '/'.join(tosca.path.split('/')[:-1])
    _log.debug('base_path: {}'.format(base_path))
    _parse_functions(tosca, inputs, base_path)
    # print(helper.print_TOSCA(tosca))

    repositories = tosca.tpl.get('repositories', None)

    tosca_name = tosca.input_path.split('/')[-1][:-5]
    tpl = Template(tosca_name)

    if hasattr(tosca, 'nodetemplates'):
        if tosca.outputs:
            tpl.outputs = tosca.outputs
        if tosca.nodetemplates:

            for node in tosca.nodetemplates:
                tpl.push(_parse_conf(tpl, node,
                                     repositories,
                                     base_path))

            _add_pointer(tpl)
            _add_back_links(tpl)
            _add_extension(tpl)

    return tpl


def _add_pointer(tpl):
    for node in tpl.nodes:
        for rel in node.relationships:
            rel.to = tpl[rel.to]


def _add_back_links(tpl):
    for node in tpl.nodes:
        for rel in node.relationships:
            rel.to.up_requirements.append(rel)


# - add pointer host_container pointer on software
# - add pointer on host property
# - add software links to the corrisponding container
def _add_extension(tpl):
    # Add the host_container property
    for node in tpl.software:
        def find_container(node):
            if isinstance(node, Container):
                return node
            elif node.host_container is not None:
                return node.host_container
            elif node.host is None:
                raise ValueError('Software component must have the "host"'
                                'requirements')
            else:
                return find_container(node.host.to)

        node.host_container = find_container(node)
        _log.debug('{} .host {}, .host_container {}'
                   ''.format(node, node.host.to, node.host_container))

    # Manage the case when a Software is connected
    # to a Container or a Software
    for node in tpl.software:
        for con in node._connection:
            if isinstance(con.to, Container):
                container = con.to
            if isinstance(con.to, Software):
                container = con.to.host_container
            _log.debug('mange connection of {} to {}'.format(node, container))
            node.host_container.add_overlay(container, con.to.name)

    # Manage the case whene a Container is connected to a Software
    for node in tpl.containers:
        for con in node._connection:
                if isinstance(con.to, Software):
                    con.alias = con.to.name
                    con.to = con.to.host_container


# def _parse_functions(tosca, inputs, base_path):
#     # Get the template nodes
#     tpl = tosca.topology_template.tpl['node_templates']
#
#     # Get the inputs of the TOSCA file
#     if 'inputs' in tosca.topology_template.tpl:
#         tosca_inputs = tosca.topology_template.tpl['inputs']
#
#     # Recurvive function searching for TOSCA function in the node
#     def parse_node(name, node):
#
#         # This function return the result of the TOSCA function
#         def execute_function(value, args):
#             if 'SELF' == args[0]:
#                 args[0] = name
#             return helper.get_attributes(args[1:], tpl[args[0]][value])
#
#         for k, v in node.items():
#             # If the function is already parse by toscaparser use the result
#             if isinstance(v, toscaparser.functions.Function):
#                 node[k] = v.result()
#             elif type(v) is dict:
#                 # Found a get_property function
#                 if 'get_property' == v:
#                     node[k] = execute_function('properties',
#                                                v['get_property'])
#                 # Found a get_artifact function
#                 if 'get_artifact' == v:
#                     art = execute_function('artifacts', v['get_artifact'])
#                     node[k] = _parse_path(base_path, art)
#                 # Found a get_input function
#                 elif 'get_input' == v:
#                     if v['get_input'] in inputs:
#                         node[k] = inputs[v['get_input']]
#                     else:
#                         node[k] = tosca_inputs[v['get_input']]['default']
#                 else:
#                     parse_node(name, v)
#
#     # Scan each component of the application to find TOSCA funcions
#     for k, v in tpl.items():
#         parse_node(k, v)


def _parse_functions(tosca, inputs, base_path):
    tpl = tosca.topology_template.tpl['node_templates']
    if 'inputs' in tosca.topology_template.tpl:
        tosca_inputs = tosca.topology_template.tpl['inputs']

    def parse_node(name, node):
        for k, v in node.items():
            # If the function is already parsed by toscaparser,
            # then use the result
            if isinstance(v, toscaparser.functions.Function):
                node[k] = v.result()
            elif isinstance(v, dict):
                # Found a get_property function
                if 'get_property' in v:
                    node[k] = get(name, 'properties', v['get_property'])
                # Found a get_artifact function
                elif 'get_artifact' in v:
                    art = get(name, 'artifacts', v['get_artifact'])
                    node[k] = _get_file(base_path, None, art)
                # Found a get_input function
                elif 'get_input' in v:
                    if v['get_input'] in inputs:
                        node[k] = inputs[v['get_input']]
                    else:
                        node[k] = tosca_inputs[v['get_input']]['default']
                else:
                    parse_node(name, v)

    # This function returns the result of the TOSCA function
    def get(name, value, args):
        if 'SELF' == args[0]:
            args[0] = name
        return helper.get_attributes(args[1:], tpl[args[0]][value])

    for k, v in tpl.items():
        parse_node(k, v)

    if 'outputs' in tosca.topology_template.tpl:
        for k, v in tosca.topology_template.tpl['outputs'].items():
            parse_node(k, v)



'''
Artifacts module
'''

class Artifact(object):

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return 'Artifact'

#
# def _get_str_name(obj):
#     return obj if isinstance(obj, six.string_types) else obj.name


class File(Artifact):

    def __init__(self, name, abs_path):
        super(self.__class__, self).__init__(name)
        split_path = abs_path.split('/')
        self.path = '/'.join(split_path[:-1])
        self.file = split_path[-1]

    @property
    def file_path(self):
        return self.path + '/' + self.file

    @property
    def format(self):
        pass

    def __str__(self):
        return 'File'


class DockerImage(Artifact):

    def __init__(self, attr):
        super(DockerImage, self).__init__('')
        self.name, self.tag = attr.split(':') if ':' in attr \
            else (attr, 'latest')

    @property
    def format(self):
        return '{}:{}'.format(self.name, self.tag)

    def __str__(self):
        return 'DockerImage'


class DockerImageExecutable(DockerImage):

    def __init__(self, name):
        super(DockerImageExecutable, self).__init__(name)

    def __str__(self):
        return 'DockerImageExecutable'


class Dockerfile(Artifact):

    def __init__(self, attr, dockerfile):
        super(Dockerfile, self).__init__('')
        self.name, self.tag = attr.split(':') if ':' in attr \
            else (attr, 'latest')
        self.dockerfile = dockerfile

    @property
    def format(self):
        return '{}:{}'.format(self.name, self.tag)

    def __str__(self):
        return 'Dockerfile'


class DockerfileExecutable(Dockerfile):

    def __init__(self, name, dockerfile):
        super(DockerfileExecutable, self).__init__(name, dockerfile)

    def __str__(self):
        return 'DockerfileExecutable'



'''
UI module
'''
import re
from os import path
from sys import argv

from six import print_

from . import __version__, helper
from .orchestrator import Orchestrator


def _usage():
    return '''
Usage: tosker FILE [COMPONENTS...] COMMAND...  [OPTIONS] [INPUTS]
       tosker ls [APPLICATION] [FILTES]
       tosker log COMPONET COMMAND
       tosker -h|--help
       tosker -v|--version
Orchestrate TOSCA applications on top of Docker.

FILE: TOSCA YAML file or CSAR file

COMMAND:
  create   Create application components
  start    Start applications components
  stop     Stop application components
  delete   Delete application components (except volume)

COMPONENTS: list of components to deploy

OPTIONS:
  -h --help      Print usage
  -q --quiet     Enable quiet mode
  --debug        Enable debugging mode (override quiet mode)
  -v --version   Print version

INPUTS: provide TOSCA inputs (syntax: --NAME VALUE)

APPLICATION: the application name (CSAR or YAML file without the extension)

FILTER:
  --name <name>    filter by the component name
  --state <state>  filter by the state (created, started, deleted)
  --type <type>    filter by the type (Container, Volume, Software)

COMPONENT: the component fullname (i.e. my_app.my_component)

Examples:
  tosker hello.yaml create start --name mario
  tosker hello.yaml stop delete -q
  tosker hello.yaml database api create start

  tosker ls
  tosker ls hello
  tosker ls hello --type Software --state started

  tosker log my_app.my_component create
'''


_FLAG = {
    '--debug': 'debug',
    '-q': 'quiet',
    '--quiet': 'quiet',
    '--help': 'help',
    '-h': 'help',
    '-v': 'version',
    '--version': 'version'
}

_CMD = {'create', 'start', 'stop', 'delete'}


def _parse_unix_input(args):
    inputs = {}
    cmds = []
    comps = []
    flags = {}
    file = ''
    mod = ''
    error = None
    p1 = re.compile('--.*')
    p2 = re.compile('-.?')
    i = 0
    while i < len(args):
        if p1.match(args[i]):
            if _FLAG.get(args[i], False):
                flags[_FLAG[args[i]]] = True
            elif (i + 1 < len(args) and
                  not p1.match(args[i + 1]) and
                  not p2.match(args[i + 1])):
                inputs[args[i][2:]] = args[i + 1]
                i += 2
                continue
            else:
                error = 'missing input value for {}'.format(args[i])
        elif p2.match(args[i]):
            if _FLAG.get(args[i], False):
                flags[_FLAG[args[i]]] = True
            else:
                error = 'known parameter.'
        elif mod == 'deploy':
            if args[i] in _CMD:
                cmds.append(args[i])
            elif len(cmds) == 0:
                comps.append(args[i])
            else:
                error = '{} is not a valid command.'.format(args[i])
        elif mod == 'ls' or mod == 'log':
            comps.append(args[i])
        elif i == 0:
            file = _check_file(args[i])
            if file is not None:
                mod = 'deploy'
            elif args[i] == 'ls':
                mod = 'ls'
            elif args[i] == 'log':
                mod = 'log'
            else:
                error = ('first argument must be a TOSCA yaml file, a CSAR or '
                         'a ZIP archive.')
        i += 1
    return error, mod, file, cmds, comps, flags, inputs


def _check_file(file):
    file_name = None
    if path.isfile(file):
        if file.endswith(('.yaml', '.csar', '.CSAR', '.YAML')):
            file_name = file
    return file_name


def _error(*str):
    print_('ERROR:', *str)
    exit(-1)


def run():
    if len(argv) < 2:
        _error('few arguments.', '\n', _usage())

    error, mod, file, cmds, comps, flags, inputs = _parse_unix_input(argv[1:])

    if error is not None:
        _error(error)

    if flags.get('help', False):
        print_(_usage())
        exit()
    elif flags.get('version', False):
        print_('TosKer version {}'.format(__version__))
        exit()

    if flags.get('debug', False):
        orchestrator = Orchestrator(log_handler=helper.get_consol_handler(),
                                    quiet=False)
    else:
        orchestrator = Orchestrator(quiet=flags.get('quiet', False))

    if mod == 'deploy':
        orchestrator.orchestrate(file, cmds, comps, inputs)
    elif mod == 'ls':
        if len(comps) > 1:
            _error('too many arguments, ls take an "application name"'
                   'and a "state"')
        else:
            app = comps[0] if len(comps) != 0 else None
            orchestrator.ls_components(app, inputs)
    elif mod == 'log':
        if len(comps) != 2:
            _error('log take a component full name and a interface name')
            return
        comp, interface = comps
        orchestrator.log(comp, interface)



import sys
import unittest
from contextlib import contextmanager

from six import StringIO

from tosker import docker_interface as docker
from tosker.orchestrator import Orchestrator

from .test_tosca_base import Test_Orchestrator


class Test_Hello(Test_Orchestrator):

    def setUp(self):
        super(self.__class__, self).setUp()
        self.file = 'data/examples/container_share_data.yaml'

    def test(self):
        self.create()
        self.start_check_exit()
        self.stop()
        self.start_check_exit()

        with open('/tmp/test_share_tosker.log') as f:
            string = f.read()
            self.assertEqual(string, '123hello123\n')

if __name__ == '__main__':
    unittest.main()



'''
Software manager module
'''
import os
from functools import wraps
from os import path
from shutil import copy

from .. import docker_interface
from ..graph.artifacts import File
from ..graph.nodes import Software
from ..helper import Logger
from ..storage import Memory


def _get_cmd(interface, force_exec=False):
    def _get_cmd_decorator(func):
        @wraps(func)
        def func_wrapper(*args):
            assert isinstance(args[0], Software)
            cmd = Software_manager._get_cmd_args(args[0], interface)
            if cmd or force_exec:
                return func(cmd, *args)
        return func_wrapper
    return _get_cmd_decorator


class Software_manager:

    @staticmethod
    @_get_cmd('create', force_exec=True)
    def create(cmd, node):
        Software_manager._copy_files(node)
        if cmd is not None:
            docker_interface.update_container(node.host_container, cmd)

    @staticmethod
    @_get_cmd('configure')
    def configure(cmd, node):
        docker_interface.update_container(node.host_container, cmd)

    @staticmethod
    @_get_cmd('start')
    def start(cmd, node):
        _log = Logger.get(__name__)
        try:
            docker_interface.exec_cmd(node.host_container, cmd)
        except Exception as e:
            _log.debug('is not running!')
            # docker_interface.delete_container(node.host_container)
            docker_interface.create_container(node.host_container,
                                              cmd=cmd,
                                              entrypoint='',
                                              from_saved=True,
                                              force=True)
            docker_interface.start_container(node.host_container)

    @staticmethod
    @_get_cmd('stop')
    def stop(cmd, node):
        _log = Logger.get(__name__)
        if docker_interface.is_running(node.host_container):
            _log.debug('exec stop command!')
            docker_interface.exec_cmd(node.host_container, cmd)

    @staticmethod
    @_get_cmd('delete')
    def delete(cmd, node):
        _log = Logger.get(__name__)
        _log.debug('exec delete command!')
        docker_interface.update_container(node.host_container, cmd)

    @staticmethod
    def _copy_files(node):
        # generate path for the tmp folder
        tmp = path.join(node.tpl.tmp_dir,
                        node.host_container.name,
                        node.name)

        # create the folder for the software
        try:
            os.makedirs(tmp)
        except os.error:
            pass

        # copy all the interfaces scripts
        for key, value in node.interfaces.items():
            copy(value['cmd'].file_path, tmp)

        # if present copy all the artifacts
        for art in node.artifacts:
            copy(art.file_path, tmp)

    @staticmethod
    def _get_cmd_args(node, interface):
        _log = Logger.get(__name__)

        def get_inside_path(p):
            if isinstance(p, File):
                p = p.file
            return path.join('/tmp/dt/', node.name, p)

        def get_cmd():
            def get_echo(string, where):
                return 'echo {} > {}'.format(string, where)

            def set_in_state():
                if interface == 'start':
                    return get_echo(Memory.STATE.STARTED.value,
                                    get_inside_path('state')) + ';'
                return ''

            def set_out_state():
                if interface == 'create':
                    return ';' + get_echo(Memory.STATE.CREATED.value,
                                          get_inside_path('state'))
                elif interface == 'start':
                    return ';' + get_echo(Memory.STATE.CREATED.value,
                                          get_inside_path('state'))
                elif interface == 'stop':
                    return ';' + get_echo(Memory.STATE.CREATED.value,
                                          get_inside_path('state'))
                elif interface == 'delete':
                    return ';' + get_echo(Memory.STATE.DELETED.value,
                                          get_inside_path('state'))
                return ''

            return '{}sh -x {} >> {} 2>&1{}'.format(
                set_in_state(),
                get_inside_path(node.interfaces[interface]['cmd']),
                get_inside_path('{}.log'.format(interface)),
                set_out_state()
            )

        if interface not in node.interfaces:
            return None
        # _log.debug('interface: {}'
        #            ''.format(node.interfaces[interface]))
        args = []
        args_env = []
        res = None
        if 'inputs' in node.interfaces[interface]:
            for key, value in node.interfaces[interface]['inputs'].items():
                if isinstance(value, File):
                    value = get_inside_path(value)
                args.append('--{} {}'.format(key, value))
                args_env.append(
                    'export INPUT_{}={}'.format(key.upper(), value))

            res = "sh -c '{};{}'".format(
                ';'.join(args_env),
                get_cmd()  # ,
                # ' '.join(args)
            )
        else:
            res = "sh -c '{}'".format(get_cmd())

        _log.debug('{} command ({}) on container {}'
                   ''.format(interface, res,
                             node.host_container))
        return res
